#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import operator
import typing
from dataclasses import dataclass, field
from types import UnionType
from typing import Any, ClassVar, Iterable

from dict_caster.extras import is_iterable
from http_tools import IncomingRequest
from http_tools.mime_types import ContentType
from http_tools.multipart_form_data import BodyPart
from init_helpers import raise_if
from init_helpers import NoValue, FieldErrors
from multidict import CIMultiDictProxy

from ..example import BaseExample
from .aggregation.call_template import CallTemplate
from .body.body_parameter import BodyParameter
from .header_parameter import HeaderParameter
from .parameter_style import ParameterStyle

T = typing.TypeVar('T')


@dataclass(frozen=True)
class ParsedBodyPart:
    filename: str | None
    header_name_to_value: dict[str, Any]
    payload: Any
    encoding: str | None
    content_type: ContentType | None


class MultipartParameterHeaders:
    def __init__(self, parent: 'MultipartParameter') -> None:
        self.parent = parent
        self._parent_header_call = CallTemplate(operator.attrgetter("header_name_to_value"), self.parent)

    def __getitem__(self, item: str):
        raise_if(not isinstance(item, str), TypeError(f'Header names must be string, got: {type(item)}: {item}'))
        return CallTemplate(operator.itemgetter(item), self._parent_header_call)

    def __str__(self):
        return f'MultipartParameterHeaders(parent={self.parent})'

    __repr__ = __str__


@dataclass(frozen=True, kw_only=True)
class MultipartParameter(BodyParameter):
    body_mime_type: ClassVar[ContentType] = ContentType.MultipartForm
    schema: type | UnionType
    allowed_content_type: ContentType | tuple[ContentType, ...] | None = None
    headers: tuple[HeaderParameter, ...] = field(init=False)
    unpack: bool = True
    explode: bool | None = None
    raw: bool = False

    def __hash__(self):
        # TODO: discover why this helps
        return hash((
            self.name, self.schema, self.is_optional, self.default, self.description, self.examples,
            self.allowed_content_type, self.headers, self.unpack, self.explode, self.is_filename_required, self.style,
        ))

    def __init__(
            self,
            name: str,
            schema: type | UnionType = bytes,
            is_optional: bool = False,
            default: Any = NoValue,
            description: str | NoValue = NoValue,
            examples: tuple[BaseExample, ...] | NoValue = NoValue,
            allowed_content_type: ContentType | Iterable[ContentType] | None = None,
            headers: Iterable[HeaderParameter] | None = None,
            unpack: bool = True,
            explode: bool | None = None,
            is_filename_required: bool = False,
            style: ParameterStyle | None = None,
            raw: bool = False,
    ) -> None:
        if not unpack and default is not NoValue:
            raise_if(not isinstance(default, BodyPart),
                     TypeError(f'With unpack=False default MUST be BodyPart, got {default=}'))
        if explode and typing.get_origin(schema) not in (list, tuple, set):
            raise TypeError(f'explode=True can be used only with iterable schema, got {schema=}')
        object.__setattr__(self, 'name', name)
        object.__setattr__(self, 'schema', schema)
        object.__setattr__(self, 'is_optional', is_optional)
        object.__setattr__(self, 'default', default)
        object.__setattr__(self, 'description', description)
        object.__setattr__(self, 'examples', examples)
        object.__setattr__(self, 'raw', raw)

        if allowed_content_type:
            if is_iterable(allowed_content_type):
                # noinspection PyUnresolvedReferences
                types_tuple = tuple(t.value for t in allowed_content_type)
                object.__setattr__(self, 'allowed_content_type', types_tuple)
                object.__setattr__(
                    self, '_check_content_type',
                    lambda x: raise_if(x not in types_tuple, KeyError(f'Expected {types_tuple}, got: {x!r}')))
            else:
                object.__setattr__(self, 'allowed_content_type', allowed_content_type)
                object.__setattr__(
                    self, '_check_content_type',
                    lambda x: raise_if(x != allowed_content_type, KeyError(f'Expected {allowed_content_type}, got: {x}')))
        else:
            object.__setattr__(self, 'allowed_content_type', None)

        if headers:
            object.__setattr__(self, 'headers', headers_tuple := tuple(headers))
            raise_if(
                len({h.name for h in headers_tuple}) < len(headers_tuple),
                KeyError('Got header parameters with ')
            )
            # object.__setattr__(
            #     self, '_check_headers',
            #     lambda x: raise_if(x not in content_type, KeyError(f'Expected {content_type}, got: {x}')))
        else:
            object.__setattr__(self, 'headers', tuple())
        object.__setattr__(self, 'unpack', unpack)
        object.__setattr__(self, 'explode', explode)
        object.__setattr__(self, 'is_filename_required', is_filename_required)
        object.__setattr__(self, 'style', style)

    def _check_content_type(self, content_type: str | None) -> None:
        pass

    def _check_headers(self, header_name_to_value: CIMultiDictProxy[str]) -> None:
        pass

    def _get(self, incoming_request: IncomingRequest) -> BodyPart | Any:
        raise_if(incoming_request.parsed_body is None, KeyError(self.name))
        raise_if(not isinstance(incoming_request.parsed_body, CIMultiDictProxy),
                 TypeError('Multipart parameter got non multipart body'))
        body: CIMultiDictProxy = incoming_request.parsed_body
        result = body.getall(self.name)
        for body_part in result:
            self._check_content_type(body_part.content_type)
        if not self.explode:
            if len(result) != 1:
                raise ValueError(f"Multipart got {len(result)} values, expected 1")
            result = result[0]

        return result

    def _extract_headers(self, value: BodyPart) -> dict[str, Any]:
        result: dict[str, Any] = {}
        field_name_to_error: dict[str, Exception] = {}
        for header in self.headers:
            if (val := value.get_all_header_values(header.name, header.default)) == NoValue and header.is_required:
                field_name_to_error[header.name] = KeyError("Missing multipart header")
                continue
            if not header.explode:
                if len(val) != 1:
                    field_name_to_error[header.name] = ValueError(f"Multipart header got {len(val)} values, expected 1")
                    continue
                val = val[0]
            result[header.name] = header._cast_value_to_schema(val)

        if field_name_to_error:
            raise FieldErrors(field_name_to_error)
        return result

    def _cast_single_value(self, value: BodyPart, schema: type[T]) -> ParsedBodyPart:
        if self.unpack:
            if isinstance(value, BodyPart):
                value = value.parse()
            return super()._cast_single_value(value=value, schema=schema)
        if not isinstance(value, BodyPart):
            raise TypeError(f'UNEXPECTED: Non unpack cast got non BodyPart {value=}')
        return ParsedBodyPart(
            filename=value.filename,
            header_name_to_value=self._extract_headers(value),
            payload=super()._cast_single_value(value=value.content if self.raw else value.parse(), schema=schema),
            encoding=value.encoding,
            content_type=value.content_type
        )

    def _cast_value_to_schema(self, value: Any):
        if self.unpack:
            if isinstance(value, BodyPart):
                return super()._cast_value_to_schema(value=value.parse())
            return super()._cast_value_to_schema(value=value)
        return super()._cast_value_to_schema(value=value)

    @property
    def payload(self) -> 'CallTemplate[bytes]':
        return CallTemplate(operator.attrgetter("payload"), self)

    @property
    def filename(self) -> 'CallTemplate[str | None]':
        return CallTemplate(operator.attrgetter("filename"), self)

    @property
    def content_type(self) -> 'CallTemplate[ContentType | None]':
        return CallTemplate(operator.attrgetter("content_type"), self)

    @property
    def header(self) -> MultipartParameterHeaders:
        return MultipartParameterHeaders(self)



from .server import gen_query_endpoint, gen_json_endpoint, get_securities_arg_names, OpenApiServer, Endpoint

from .spec import BaseSchema, TypeSchema, SchemaBuilder
from .spec import CompositeSchema, AllOfSchema, AnyOfSchema, EmptySchema
from .spec import ArraySchema, ObjectSchema
from .spec import BooleanSchema, IntegerSchema, NullSchema, NumberSchema, StringSchema
from .spec import SchemaFormat, StringSchemaFormat, NumberSchemaFormat, IntegerSchemaFormat

from .spec import BodyPartDescription
from .spec import ReferencableResource
from .spec import SpecRef
from .spec import SpecResource
from .spec import SpecGenerator
from .spec import Example, UrlExample, BaseExample, AnswerExample
from .spec import Operation
from .spec import CallParameter, SpecParameter, ParameterLocation, ParameterStyle
from .spec import CallTemplate, ParameterAggregation
from .spec import BodyParameter, RawBodyParameter, JsonBodyParameter, JsonWholeBodyParameter
from .spec import HeaderParameter, PathParameter, QueryParameter, CookieParameter
from .spec import LiteralParameter, MultipartParameter, SecurityParameter
from .spec import Unauthorized, Forbidden, AuthInfoException
from .spec import SecuritySchemeType, SecurityScheme
from .spec import ApiKeySecurityScheme, HttpSecurityScheme, HttpBearerSecurityScheme, HttpBasicSecurityScheme
from .spec import OAuth2Flow, OAuth2SecurityScheme, OpenIdConnectSecurityScheme
from .spec import SecurityRequirement, get_securities_arg_names
from .spec import RequestBody
from .spec import Tag

from .freeze import freeze

# backward compatability
from .server import gen_query_endpoint as gen_query_rpc_endpoint, gen_json_endpoint as gen_json_rpc_endpoint
from .server import Endpoint as RpcEndpoint



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
import functools
import typing
from dataclasses import dataclass
from types import UnionType
from typing import Any

from frozendict import frozendict
from http_tools import IncomingRequest
from init_helpers import AnyType

from ..schema.base_schema import BaseSchema
# from ..schema.spec_schema import SpecSchema

if typing.TYPE_CHECKING:
    from .spec_parameter import SpecParameter


@dataclass(frozen=True, slots=True, kw_only=True)
class CallParameter(abc.ABC):
    """Параметр вызова процедуры, schema - тип, key - ключ при вызове"""
    schema: AnyType | BaseSchema
    # key: str

    @abc.abstractmethod
    async def get(self, incoming_request: IncomingRequest, security_kwargs: dict[str, Any]) -> Any:
        pass

    @abc.abstractmethod
    def get_spec_parameters(self) -> frozendict['SpecParameter', None]:
        pass

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        return tuple([f"schema={self.schema.__name__ if isinstance(self.schema, type) else self.schema}"])

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass
from typing import ClassVar, Any

from http_tools import ContentType, IncomingRequest
from init_helpers import raise_if

from .body_parameter import BodyParameter


@dataclass(frozen=True, kw_only=True)
class JsonBodyParameter(BodyParameter):
    body_mime_type: ClassVar[ContentType] = ContentType.Json

    def _get(self, incoming_request: IncomingRequest) -> Any:
        raise_if(incoming_request.parsed_body is None, KeyError(self.name))
        return incoming_request.parsed_body[self.name]



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from abc import ABC
from dataclasses import dataclass
from typing import Generic, ClassVar, Mapping

from frozendict import frozendict
from init_helpers import Jsonable

from openapi_tools.spec import SpecResource, SpecRef
from .security_scheme import SecurityScheme, AuthInfo
from .security_scheme_type import SecuritySchemeType


@dataclass(frozen=True)
class OAuth2SecurityScheme(SecurityScheme, Generic[AuthInfo], ABC):
    type_: ClassVar[SecuritySchemeType] = SecuritySchemeType.oauth2
    # flow_to_info: str ???

    def get_spec_dict(self, dependency_to_ref: Mapping[SpecResource, SpecRef]) -> frozendict[str, Jsonable]:
        raise NotImplementedError



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
import functools
from dataclasses import dataclass
from types import UnionType
from typing import Any, Annotated, Mapping, get_args, ClassVar, Iterable, TypeVar

from frozendict import frozendict
from http_tools import IncomingRequest, ContentType
from init_helpers import append_if, Jsonable
from init_helpers import NoValue, convert_to_type

from .call_parameter import CallParameter
from .parameter_location import ParameterLocation
from .parameter_style import ParameterStyle
from ..example import BaseExample, get_examples_as_dict
from ..schema.base_schema import BaseSchema
from ..spec_ref import SpecRef
from ..spec_resource import SpecResource
from ...freeze import freeze


T = TypeVar('T')


@dataclass(frozen=True, slots=True, kw_only=True, repr=False)
class SpecParameter(CallParameter, SpecResource, abc.ABC):
    """Параметр, входящий в спецификацию, location - расположение(хедер/query/path/т.п), name - имя в расположении"""
    # schema_builder: ClassVar[Callable[[AnyType], BaseSchema]] = SchemaBuilder.build_schema
    location: ClassVar[ParameterLocation]
    name: str
    is_optional: bool = False
    default: Any = NoValue
    description: str = ''

    examples: Iterable[BaseExample] = tuple()

    explode: bool | None = None
    style: ParameterStyle | None = None

    content_type: ContentType | None = None

    def __post_init__(self):
        if self.default is not NoValue:
            object.__setattr__(self, 'default', freeze(self.default))
        object.__setattr__(self, 'examples', tuple(self.examples))
        if self.content_type is not None and (self.explode is not None or self.style is not None):
            raise ValueError(f"{self} HAS {self.content_type=}, so it MUST have style=None an explode=None")

    @property
    def _schema(self) -> BaseSchema:
        return self.schema
        # return self.schema if isinstance(self.schema, BaseSchema) else self.schema_builder(self.schema)

    def get_spec_dependencies(self) -> frozenset[SpecResource]:
        result: set[SpecResource] = {self._schema}
        result = result.union(self.examples)
        return frozenset(result)

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef]) -> frozendict[str, Jsonable]:
        result = {'name': self.name, 'in': self.location}
        result |= {'description': self.description} if self.description else {}

        rendered_schema = dependency_to_ref[self._schema]
        if self.content_type:
            if self.explode is not None or self.style is not None:
                raise ValueError(f"{self} HAS {self.content_type=}, so it MUST have style=None an explode=None")
            result['content'] = {self.content_type: rendered_schema}
        else:
            result['schema'] = rendered_schema

        result |= {'style': self.style} if self.style else {}
        result |= {'required': self.is_required} if self.is_required else {}
        result |= {'default': self.default} if self.default is not NoValue else {}
        result |= get_examples_as_dict(self.examples, dependency_to_ref)
        return frozendict(result)

    @property
    def annotated_schema(self) -> type | UnionType | Annotated:
        notes = []
        append_if(self.description is not NoValue, notes, self.description)
        append_if(self.examples is not NoValue, notes, *self.examples if self.examples is not NoValue else [])
        return Annotated[self.schema, *notes] if notes else self.schema

    @property
    def is_required(self) -> bool:
        return self.default is NoValue and not self.is_optional and self.location != ParameterLocation.path

    def get_spec_parameters(self) -> frozendict['SpecParameter', None]:
        return frozendict({self: None})

    def _cast_value_to_schema(self, value: Any):
        if self.explode:
            type_args = get_args(self.schema)
            if len(type_args) != 1:
                raise TypeError(f'Only generic containers with single argument (list, set, simple tuple) are supported')
            type_arg = type_args[0]
            return [self._cast_single_value(value=val, schema=type_arg) for val in value]
        return self._cast_single_value(value=value, schema=self.schema)

    def _cast_single_value(self, value: Any, schema: type[T]) -> T:
        if self.style is not None and isinstance(value, str):
            value = self._apply_style(value)
        return convert_to_type(value=value, type_=schema)

    async def get(self, incoming_request: IncomingRequest, security_kwargs: dict[str, Any]) -> Any | NoValue:
        try:
            result = self._get(incoming_request)
        except KeyError:
            if self.default is NoValue:
                if self.is_optional:
                    return NoValue
                raise
            result = self.default
        return self._cast_value_to_schema(result)

    @abc.abstractmethod
    def _get(self, incoming_request: IncomingRequest) -> Any:
        pass

    def _apply_style(self, value: str) -> Any:
        if self.style == ParameterStyle.simple:
            assert isinstance(value, str), "Expected string value, got:"
            return value.split(",")
        return value

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        parts = list(CallParameter._get_repr_parts(self))
        parts.append(f'name={self.name!r}')
        parts += [f'is_optional={self.is_optional!r}'] if self.is_optional else []
        parts += [f'default={self.default!r}'] if self.default is not NoValue else []
        parts += [f'description={self.description!r}'] if self.description else []
        parts += [f'examples={self.examples!r}'] if self.examples else []
        parts += [f'explode={self.explode!r}'] if self.explode is not None else []
        parts += [f'style={self.style!r}'] if self.style is not None else []
        parts += [f'content_type={self.content_type!r}'] if self.content_type is not None else []
        return tuple(parts)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from .schema import BaseSchema, TypeSchema, SchemaBuilder
from .schema import CompositeSchema, AllOfSchema, AnyOfSchema, EmptySchema
from .schema import ArraySchema, ObjectSchema
from .schema import BooleanSchema, IntegerSchema, NullSchema, NumberSchema, StringSchema
from .schema import SchemaFormat, StringSchemaFormat, NumberSchemaFormat, IntegerSchemaFormat

from .content_schema import ContentSchema
from .body_part_description import BodyPartDescription
from .referencable_resource import ReferencableResource
from .spec_ref import SpecRef
from .spec_resource import SpecResource
from .spec_generator import SpecGenerator
from .operation import Operation
from .example import Example, UrlExample, BaseExample, AnswerExample

from .parameter import CallParameter, SpecParameter, ParameterLocation, ParameterStyle
from .parameter import CallTemplate, ParameterAggregation
from .parameter import BodyParameter, RawBodyParameter, JsonBodyParameter, JsonWholeBodyParameter
from .parameter import HeaderParameter, PathParameter, QueryParameter, CookieParameter
from .parameter import LiteralParameter, MultipartParameter, SecurityParameter

from .security import Unauthorized, Forbidden, AuthInfoException, resolver_cache
from .security import SecuritySchemeType, SecurityScheme
from .security import ApiKeySecurityScheme, HttpSecurityScheme, HttpBearerSecurityScheme, HttpBasicSecurityScheme
from .security import OAuth2Flow, OAuth2SecurityScheme, OpenIdConnectSecurityScheme
from .security import SecurityRequirement, get_securities_arg_names
from .request_body import RequestBody
from .media_type import MediaType
from .response import Response
from .tag import Tag



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
from dataclasses import dataclass
from typing import ClassVar

from http_tools import ContentType

from ..parameter_location import ParameterLocation
from ..spec_parameter import SpecParameter


@dataclass(frozen=True, kw_only=True)
class BodyParameter(SpecParameter, abc.ABC):
    location: ClassVar[ParameterLocation] = ParameterLocation.body
    body_mime_type: ClassVar[ContentType] = None



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
import re
from dataclasses import dataclass
from typing import ClassVar, Mapping

from frozendict import frozendict
from init_helpers import Jsonable

from openapi_tools.spec.spec_resource import SpecResource, SpecRef
from .schema_format import StringSchemaFormat
from .type_schema import TypeSchema
from ..example import Example


@dataclass(frozen=True, slots=True)
class StringSchema(TypeSchema):
    type_: ClassVar[str] = "string"
    format_: StringSchemaFormat | None = None
    min_length: int | None = None
    max_length: int | None = None
    pattern: str = ''

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef | dict]) -> frozendict[str, Jsonable]:
        result = {}
        if self.pattern:
            result['pattern'] = self.pattern
            if self.examples:
                if any(not (re.match(self.pattern, v := e.value)) for e in self.examples if isinstance(e, Example)):
                    raise ValueError(f'Example value: {v!r} does not match pattern: {self.pattern!r}')
        if self.min_length is not None:
            result |= {'minLength': self.min_length}
        if self.max_length is not None:
            result |= {'maxLength': self.max_length}
        return frozendict(TypeSchema.get_spec_dict(self, dependency_to_ref) | result)

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        parts = list(TypeSchema._get_repr_parts(self))
        parts += [f'min_length={self.min_length!r}'] if self.min_length else []
        parts += [f'max_length={self.max_length!r}'] if self.max_length else []
        return tuple(parts)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
from dataclasses import dataclass, field
from typing import ClassVar, Mapping

from frozendict import frozendict
from http_tools import ContentType
from init_helpers import Jsonable

from .spec_resource import SpecResource
from .spec_ref import SpecRef
from .maybe_referencable_resource import MaybeReferencableResource
from .media_type import MediaType


@dataclass(frozen=True, slots=True)
class Response(MaybeReferencableResource):
    spec_path: ClassVar[list[str]] = ['components', 'responses']
    description: str = ''
    # headers: Mapping[str, str] = field(default_factory=frozendict)  # todo: implement headers
    content_type_to_media_type: Mapping[ContentType, MediaType] = field(default_factory=frozendict)
    # links: Mapping[str, Link]  # todo: implement links
    key: str | None = field(default=None, kw_only=True)

    def __post_init__(self):
        object.__setattr__(self, 'content_type_to_media_type', frozendict(self.content_type_to_media_type))

    def get_key(self) -> str:
        return None
        return self.key

    def get_spec_dependencies(self) -> frozenset['SpecResource']:
        return frozenset(self.content_type_to_media_type.values())

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef]) -> frozendict[str, Jsonable]:
        result = {"description": self.description}
        if self.content_type_to_media_type:
            result["content"] = {t: dependency_to_ref[s] for t, s in self.content_type_to_media_type.items()}

        return frozendict(result)

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        result = ([repr({ct.value: mt for ct, mt in self.content_type_to_media_type.items()})]
                  if self.content_type_to_media_type else [])
        result += [f'key={self.key!r}'] if self.key else []
        result += [f'description={self.description!r}'] if self.description else []
        return tuple(result)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import asyncio
from dataclasses import dataclass, field
from typing import Generic, Callable, TypeVarTuple, Unpack, Any, Awaitable, Literal, \
    get_origin, get_args
from typing import Mapping, Iterable

from async_tools import acall
from frozendict import frozendict
from http_tools import IncomingRequest
from init_helpers import Jsonable, raise_if

from openapi_tools.spec import SpecResource
from ..exceptions import AuthInfoException
from ..scheme import SecurityScheme, AuthInfo

Ts = TypeVarTuple('Ts')


@dataclass(frozen=True, slots=True)
class SecurityRequirement(SpecResource, Generic[Unpack[Ts]]):
    scheme_to_scopes: Mapping[SecurityScheme, frozenset[str]] = field(default_factory=frozendict)
    argument_name_to_getter: dict[str, Callable[[Unpack[Ts]], Any]] = field(default_factory=dict, hash=False)
    string_key: str = field(init=False)
    do_log: bool = field(init=False)

    def __init__(self, scheme_to_scopes: Mapping[SecurityScheme, Iterable[str]] | None = None,
                 argument_name_to_getter: Mapping[str, Callable[[AuthInfo], Awaitable | Any]] | None = None):
        object.__setattr__(
            self, 'scheme_to_scopes',
            frozendict({scheme: frozenset(scopes) for scheme, scopes in (scheme_to_scopes or {}).items()}))
        object.__setattr__(self, 'argument_name_to_getter', frozendict(argument_name_to_getter or {}))
        object.__setattr__(self, 'string_key', ';'.join(
            f'{scheme.get_key()}[{",".join(sorted(scopes))}]' for scheme, scopes in self.scheme_to_scopes.items()
        ))
        object.__setattr__(self, 'do_log', any(s.do_log for s in self.scheme_to_scopes))

    def get_spec_dependencies(self) -> frozenset['SpecResource']:
        return frozenset(self.scheme_to_scopes.keys())

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', str]) -> frozendict[str, Jsonable]:
        # noinspection PyTypeChecker
        return frozendict({dependency_to_ref[scheme]: scopes for scheme, scopes in self.scheme_to_scopes.items()})

    async def evaluate(self, incoming_request: IncomingRequest) -> dict:
        async with asyncio.TaskGroup() as task_group:
            tasks = [
                task_group.create_task(scheme.evaluate(incoming_request, scopes))
                for scheme, scopes in self.scheme_to_scopes.items()
            ]
        auth_infos = [t.result() for t in tasks]
        try:
            return {name: await acall(getter(*auth_infos)) for name, getter in self.argument_name_to_getter.items()}
        except (KeyError, TypeError, ValueError) as e:
            raise AuthInfoException(*e.args) from e

    def use(self, **kwargs: type[Literal] | Callable[[Unpack[Ts]], Any]) -> 'SecurityRequirement':
        argument_name_to_getter = dict(self.argument_name_to_getter)
        for key, value in kwargs.items():
            if get_origin(value) is Literal:
                args_args = get_args(value)
                raise_if(len(args_args) != 1, TypeError(f'Only single literal values allowed, got: {args_args!r}'))
                argument_name_to_getter[key] = lambda *a: args_args[0]
            elif callable(value):
                argument_name_to_getter[key] = value
            else:
                raise TypeError(f'Got unexpected value: {value}, Allowed: Literal/Callable[[AuthInfo], Any]')
        return SecurityRequirement(scheme_to_scopes=self.scheme_to_scopes,
                                   argument_name_to_getter=argument_name_to_getter)

    def has(self, **kwargs: type[Literal] | Callable[[Unpack[Ts]], Any]) -> 'SecurityRequirement':
        return self.use(**kwargs)

    def _repr(self) -> list[str]:
        parts = []
        if self.scheme_to_scopes:
            parts.append(f'scheme_to_scopes={ {scheme: tuple(scopes) for scheme, scopes in self.scheme_to_scopes.items()} }')
        return parts

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._repr())})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
from dataclasses import dataclass
from typing import Mapping, ClassVar

from frozendict import frozendict
from init_helpers import Jsonable

from openapi_tools.spec.spec_resource import SpecResource, SpecRef
from .base_schema import BaseSchema
from .type_schema import TypeSchema


@dataclass(frozen=True, slots=True)
class ArraySchema(TypeSchema):
    type_: ClassVar[str] = "array"
    items_schema: BaseSchema | SpecRef | None = None
    is_uniqueness_required: bool = False

    @functools.cache
    def get_spec_dependencies(self) -> frozenset['SpecResource']:
        return frozenset(BaseSchema.get_spec_dependencies(self) | (
            {self.items_schema} if isinstance(self.items_schema, SpecResource) else set()))

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef | dict]) -> frozendict[str, Jsonable]:
        result = {}
        if (sub_schema := self.items_schema) is not None:
            result |= {'items': sub_schema if isinstance(sub_schema, SpecRef) else dependency_to_ref[sub_schema]}
        if self.is_uniqueness_required:
            result |= {'uniqueItems': self.is_uniqueness_required}
        return frozendict(TypeSchema.get_spec_dict(self, dependency_to_ref) | result)

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        parts = list(TypeSchema._get_repr_parts(self))
        parts += [f'items_schema={self.items_schema!r}'] if self.items_schema else []
        parts += [f'is_uniqueness_required={self.is_uniqueness_required!r}'] if self.is_uniqueness_required else []
        return tuple(parts)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass
from functools import cached_property
from typing import Mapping, ClassVar, Generic

from frozendict import frozendict
from http_tools import IncomingRequest
from init_helpers import Jsonable

from ...parameter import SpecParameter
from ...spec_ref import SpecRef
from ...spec_resource import SpecResource
from openapi_tools.spec.parameter import ParameterLocation, QueryParameter, PathParameter, HeaderParameter, CookieParameter
from .security_scheme import SecurityScheme, AuthInfo, AuthToken
from .security_scheme_type import SecuritySchemeType
from ..exceptions import Unauthorized


@dataclass(frozen=True)
class ApiKeySecurityScheme(SecurityScheme, Generic[AuthInfo]):
    type_: ClassVar[SecuritySchemeType] = SecuritySchemeType.api_key
    location: ParameterLocation
    name: str

    def get_spec_dict(self, dependency_to_ref: Mapping[SpecResource, SpecRef]) -> frozendict[str, Jsonable]:
        return SecurityScheme.get_spec_dict(self, dependency_to_ref) | {'in': self.location, 'name': self.name}

    def get_key(self) -> str:
        return f'{self.type_.name}_in_{self.location.name}_{self.name}'

    def __post_init__(self):
        assert self._parameter

    @cached_property
    def _parameter(self) -> SpecParameter:
        if self.location == ParameterLocation.query:
            return QueryParameter(name=self.name, schema=str)
        if self.location == ParameterLocation.path:
            return PathParameter(name=self.name, schema=str)
        if self.location == ParameterLocation.header:
            return HeaderParameter(name=self.name, schema=str)
        elif self.location == ParameterLocation.cookie:
            return CookieParameter(name=self.name, schema=str)
        raise NotImplementedError(f"ApiKeySecurityScheme does not support location {self.location!r} ")

    async def _extract_token(self, incoming_request: IncomingRequest) -> AuthToken:
        try:
            return await self._parameter.get(incoming_request, {})
        except KeyError:
            raise Unauthorized from None
